from __future__ import annotations

import sys
import aiohttp
import logging

from urllib.parse import urlencode

from . import __version__, utils
from .utils import MISSING

from typing import Any, ClassVar, Coroutine, Optional, TypeVar, TYPE_CHECKING

_log = logging.getLogger(__name__)

__all__ = ('HTTPClient',)

if TYPE_CHECKING:

    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]

    from .types.marblex import (
        Coin as CoinPayload
    )
    from .types.exchange import (
        LoremboardExchange as LoremboardExchangePayload
    )

class Route:

    BASE_MARBLEX_URL: ClassVar[str] = 'https://ninokuni.marblex.io/api'
    BASE_LOREM_URL: ClassVar[str] = 'https://api.loremboard.finance/api/v1'

    def __init__(
        self,
        method: str,
        path: str,
        endpoint: Optional[str] = "mbx",
        **parameters: Any
    ) -> None:
        self.method: str = method
        self.path: str = path

        url = ''
        if endpoint == "mbx":
            url = self.BASE_MARBLEX_URL + path
        elif endpoint == "lorem":
            url = self.BASE_LOREM_URL + path

        if parameters:
            url = url + '?' + urlencode(parameters)
        self.url: str = url

class HTTPClient:

    """Represents an HTTP client sending HTTP requests to the Marblex API."""

    def __init__(self) -> None:
        self.__session: aiohttp.ClientSession = MISSING

        user_agent = 'Marblex.py (https://github.com/staciax/marblex.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self.user_agent: str = user_agent.format(__version__, sys.version_info, aiohttp.__version__)

    async def request(self, route: Route, **kwargs: Any) -> Any:
        method = route.method
        url = route.url

        if self.__session is MISSING:
            self.__session = aiohttp.ClientSession()

        # headers
        kwargs['headers'] = {
            'User-Agent': self.user_agent,
        }

        # verify ssl
        kwargs['verify_ssl'] = False

        async with self.__session.request(method, url, **kwargs) as response:

            if response.status == 200:
                _log.debug(f'HTTP request status: {response.status}')
                data = await utils.json_or_text(response)
                return data
            else:
                _log.error(f'HTTP request failed: {response.url}')
                raise Exception(f'HTTP request failed: {response.url}')

            # TODO: error handling

    # marblex endpoints

    def fetch_territe_token(self) -> Response[CoinPayload]:
        return self.request(Route('GET', '/price', tokenType='NKT'))

    def fetch_asterite_token(self) -> Response[CoinPayload]:
        return self.request(Route('GET', '/price', tokenType='NKA'))

    def fetch_inetrium_token(self) -> Response[CoinPayload]:
        return self.request(Route('GET', '/price', tokenType='ITU'))

    # exchange endpoints

    def fetch_loremboard(self) -> Response[LoremboardExchangePayload]:
        return self.request(Route('GET', '/dashboard/fiat/latest', endpoint="lorem"))

    async def close(self) -> None:
        _log.debug('Closing HTTP client')
        await self.__session.close()
        self.__session = MISSING


"""
Marblex API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the Marblex API.
:copyright: (c) 2022-present xStacia
:license: MIT, see LICENSE for more details.
"""
__title__ = 'marblex.py'
__version__ = '1.3.2'
__author__ = 'xStacia'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present xStacia'

from .client import *
from .models import *


from __future__ import annotations

import json

from typing import Any, Dict, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from aiohttp import ClientResponse

def _to_dict(text: str) -> dict:
    """Convert text to dict"""
    return json.loads(text)

async def json_or_text(response: ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')
    if 'Content-Type' in response.headers:
        if response.headers['Content-Type'] == 'application/data':
            return await response.json()

    try:
        return _to_dict(text)
    except (json.JSONDecodeError, TypeError):
        return text

class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __hash__(self):
        return 0

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()



from __future__ import annotations

import logging

from .http import HTTPClient
from .models import *

from typing import Type, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self
    from types import TracebackType


_log = logging.getLogger(__name__)

__all__ = ('Client', 'Marblex')

class Client:

    """ Marblex API Wrapper Client """

    def __init__(self):
        self.__http = HTTPClient()
        self._nkt_percent: str = ''
        self._nka_percent: str = ''
        self._itu_percent: str = ''

        self._closed: bool = False

    async def get_territe_token(self) -> Optional[Coin]:
        """ Get NKT Coin. """
        data = await self.__http.fetch_territe_token()
        coin = Coin(client=self, data=data)
        await coin.get_exchange()
        return coin

    async def get_asterite_token(self) -> Optional[Coin]:
        """ Get NKA Coin. """
        data = await self.__http.fetch_asterite_token()
        coin = Coin(client=self, data=data)
        await coin.get_exchange()
        return coin

    async def get_inetrium_token(self) -> Optional[Coin]:
        """ Get ITU Coin. """
        data = await self.__http.fetch_inetrium_token()
        coin = Coin(client=self, data=data)
        await coin.get_exchange()
        return coin

    async def get_exchange(self) -> Optional[Exchange]:
        """ Return the Exchange object. """
        data = await self.__http.fetch_loremboard()
        return Exchange(client=self, data=data)

    async def __aenter__(self) -> Self:
        _log.info('Opened the Marblex client.')
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if not self.is_closed():
            await self.close()

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        await self.__http.close()
        _log.info('Closed the Marblex client.')

    def is_closed(self) -> bool:
        return self._closed

Marblex = Client